/**
 * @file artis/kernel/AbstractModel.hpp
 * @author See the AUTHORS file
 */

/*
 * Copyright (C) 2012-2016 ULCO http://www.univ-littoral.fr
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __ARTIS_KERNEL_ABSTRACT_ATOMIC_MODEL_HPP
#define __ARTIS_KERNEL_ABSTRACT_ATOMIC_MODEL_HPP

#include <artis/kernel/AbstractModel.hpp>
#include <artis/kernel/AbstractCoupledModel.hpp>
#include <artis/kernel/Internals.hpp>
#include <artis/kernel/States.hpp>

#include <artis/utils/DateTime.hpp>
#include <artis/utils/Exception.hpp>

namespace artis { namespace kernel {

template < typename T, typename U, typename V >
class AbstractAtomicModel : public AbstractModel < U, V >,
                            public States < T, U >,
                            public Internals < T, U >,
                            public Externals < T, U >
{
    typedef AbstractModel < U, V > type;

public:
    AbstractAtomicModel(const type* parent = 0) : type(parent)
    { }

    virtual ~AbstractAtomicModel()
    { }

    virtual void after(typename U::type t)
    {

#ifdef WITH_TRACE
        utils::Trace < utils::DoubleTime >::trace()
            << utils::TraceElement < utils::DoubleTime >("KERNEL", t,
                                                         utils::KERNEL)
            << " - AFTER - "
            << AbstractAtomicModel < T, U, V >::path(this);
        utils::Trace < utils::DoubleTime >::trace().flush();
#endif

    }

    virtual void before(typename U::type t)
    {

#ifdef WITH_TRACE
        utils::Trace < utils::DoubleTime >::trace()
            << utils::TraceElement < utils::DoubleTime >("KERNEL", t,
                                                         utils::KERNEL)
            << " - BEFORE - "
            << AbstractAtomicModel < T, U, V >::path(this);
        utils::Trace < utils::DoubleTime >::trace().flush();
#endif

    }

    virtual bool check(typename U::type t) const
    { return Externals < T, U >::check(t); }

    virtual void compute(typename U::type t, bool update) = 0;

    virtual double get(typename U::type t, unsigned int index) const
    {
        if (type::last_time != t) {
            throw utils::InvalidGet("Variable not computed");
        }
        return static_cast < const T* >(this)->*(
            Internals < T, U >::get(index));
    }

    virtual int getI(typename U::type t, unsigned int index) const
    {
        if (type::last_time != t) {
            throw utils::InvalidGet("Variable not computed");
        }
        return static_cast < const T* >(this)->*(
            Internals < T, U >::getI(index));
    }

    virtual bool getB(typename U::type t, unsigned int index) const
    {
        if (type::last_time != t) {
            throw utils::InvalidGet("Variable not computed");
        }
        return static_cast < const T* >(this)->*(
            Internals < T, U >::getB(index));
    }

    virtual void init(typename U::type t, const V& parameters) = 0;

    bool is_computed(typename U::type t, unsigned int /* index */) const
    { return type::last_time == t; }

    bool is_stable(typename U::type t) const
    { return type::last_time == t; }

    virtual bool is_updated() const
    { return Externals < T, U >::updated; }

    virtual void restore(const State < U >& state)
    {
        States < T, U >::restore(state);
        Internals < T, U >::restore(state);
        type::last_time = state.last_time();
    }

    virtual void save(State < U >& state) const
    {
        States < T, U >::save(state);
        Internals < T, U >::save(state);
        state.last_time(type::last_time);
    }

    virtual void stable()
    { Externals < T, U >::updated = false; }
};

template < typename T, typename U, typename V >
struct SA_t
{
    SA_t(double t, AbstractAtomicModel < T, U, V >* model,
         unsigned int index) : t(t), model(model), index(index)
    { }

    double t;
    AbstractAtomicModel < T, U, V >* model;
    unsigned int index;
};

template < typename T, typename U, typename V >
struct IN_SA_t : public SA_t < T, U, V >
{
    IN_SA_t(double t, AbstractAtomicModel < T, U, V >* model,
            unsigned int index) : SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
struct IN_SA_I_t : public IN_SA_t < T, U, V >
{
    IN_SA_I_t(double t, AbstractAtomicModel < T, U, V >* model,
              unsigned int index) : IN_SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
struct IN_SA_B_t : public IN_SA_t < T, U, V >
{
    IN_SA_B_t(double t, AbstractAtomicModel < T, U, V >* model,
              unsigned int index) : IN_SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
struct OUT_SA_t : public SA_t < T, U, V >
{
    OUT_SA_t(double t, AbstractAtomicModel < T, U, V >* model,
            unsigned int index) : SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
struct OUT_SA_I_t : public OUT_SA_t < T, U, V >
{
    OUT_SA_I_t(double t, AbstractAtomicModel < T, U, V >* model,
               unsigned int index) : OUT_SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
struct OUT_SA_B_t : public OUT_SA_t < T, U, V >
{
    OUT_SA_B_t(double t, AbstractAtomicModel < T, U, V >* model,
               unsigned int index) : OUT_SA_t < T, U , V >(t, model, index)
    { }
};

template < typename T, typename U, typename V >
IN_SA_t < T, U, V > IN(double t, AbstractAtomicModel < T, U, V >* model,
                       unsigned int index)
{ return IN_SA_t < T, U, V >(t, model, index); }

template < typename T, typename U, typename V >
IN_SA_I_t < T, U, V > IN_I(double t, AbstractAtomicModel < T, U, V >* model,
                           unsigned int index)
{ return IN_SA_I_t < T, U, V >(t, model, index); }

template < typename T, typename U, typename V >
IN_SA_B_t < T, U, V > IN_B(double t, AbstractAtomicModel < T, U, V >* model,
                           unsigned int index)
{ return IN_SA_B_t < T, U, V >(t, model, index); }

template < typename T, typename U, typename V >
OUT_SA_t < T, U, V > OUT(double t,
                         AbstractAtomicModel < T, U, V >* model,
                         unsigned int index)
{ return OUT_SA_t < T, U, V >(t, model, index); }

template < typename T, typename U, typename V >
OUT_SA_I_t < T, U, V > OUT_I(double t,
                             AbstractAtomicModel < T, U, V >* model,
                             unsigned int index)
{ return OUT_SA_I_t < T, U, V >(t, model, index); }

template < typename T, typename U, typename V >
OUT_SA_B_t < T, U, V > OUT_B(double t,
                             AbstractAtomicModel < T, U, V >* model,
                             unsigned int index)
{ return OUT_SA_B_t < T, U, V >(t, model, index); }

template < typename T1, typename U1, typename V1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_t < T2, U2, V2 > out, IN_SA_t < T1, U1, V1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->get(out.t, out.index));
    }
}

template < typename T1, typename U1, typename V1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_I_t < T2, U2, V2 > out, IN_SA_I_t < T1, U1, V1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->getI(out.t, out.index));
    }
}

template < typename T1, typename U1, typename V1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_B_t < T2, U2, V2 > out, IN_SA_B_t < T1, U1, V1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->getB(out.t, out.index));
    }
}

template < typename T1, typename U1, typename V1, typename W1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_t < T2, U2, V2 > out, IN_SC_t < T1, U1, V1, W1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->get(out.t, out.index));
    }
}

template < typename T1, typename U1, typename V1, typename W1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_B_t < T2, U2, V2 > out, IN_SC_B_t < T1, U1, V1, W1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->getB(out.t, out.index));
    }
}

template < typename T1, typename U1, typename V1, typename W1,
           typename T2, typename U2, typename V2 >
void operator>>(OUT_SA_I_t < T2, U2, V2 > out, IN_SC_I_t < T1, U1, V1, W1 > in)
{
    if (out.model->is_computed(out.t, out.index)) {
        in.model->put(out.t, in.index, out.model->getI(out.t, out.index));
    }
}

} }

#endif
